import datetime
import calendar

from django.conf import settings
from django.utils.timezone import get_current_timezone


def convert(other):
    if isinstance(other, bool):
        return Bool(other)
    if isinstance(other, str):
        return String(other)
    if isinstance(other, int):
        return Integer(other)
    if isinstance(other, float):
        return Float(other)
    if isinstance(other, datetime.datetime):
        return DateTime(other)
    if isinstance(other, datetime.date):
        return Date(other)
    if isinstance(other, SQLExpression):
        return other

    raise TypeError('Cannot convert type "%s"' % type(other))


class CompilableSQL(object):
    """
    An abstract method that defines the API
    to transform itself into a sql expression.
    """
    def as_sql(self):
        """
        Returns the query as an sql without parameters
        """
        raise NotImplementedError

    def get_params(self):
        """
        Returns the list parameters of the query
        """
        raise NotImplementedError('%s' % self.__class__)

    def sql(self):
        """
        Returns the sql expression with substituting parameters
        """
        return self.as_sql() % self.get_params()

    def resolve_columns(self, index):
        return self


class SQLExpression(CompilableSQL):
    """
    A general sql expression.
    """
    def __init__(self, value):
        self._value = value

    @property
    def value(self):
        return self._value

    def type(self):
        """
        The output type of the expression (e.g. `Bool`).
        It must return a subclass of `Value`.
        """
        raise NotImplementedError

    def __add__(self, other):
        other = convert(other)
        return Add(self, other)

    def __radd__(self, other):
        other = convert(other)
        return Add(other, self)

    def __sub__(self, other):
        other = convert(other)
        return Subtract(self, other)

    def __rsub__(self, other):
        other = convert(other)
        return Subtract(other, self)

    def __mul__(self, other):
        other = convert(other)
        return Multiply(self, other)

    def __rmul__(self, other):
        other = convert(other)
        return Multiply(other, self)

    def __truediv__(self, other):
        other = convert(other)
        return Divide(self, other)

    def __rtruediv__(self, other):
        other = convert(other)
        return Divide(other, self)

    def __pow__(self, other):
        other = convert(other)
        return Power(self, other)

    def __rpow__(self, other):
        other = convert(other)
        return Power(other, self)

    def __neg__(self):
        return Neg(self)

    ### operators that return bool

    def __eq__(self, other):
        other = convert(other)
        return Equal(self, other)

    def __ne__(self, other):
        other = convert(other)
        return NotEqual(self, other)

    def __ge__(self, other):
        other = convert(other)
        return GreaterEqualThan(self, other)

    def __gt__(self, other):
        other = convert(other)
        return GreaterThan(self, other)

    def __le__(self, other):
        other = convert(other)
        return LessEqualThan(self, other)

    def __lt__(self, other):
        other = convert(other)
        return LessThan(self, other)

    ### operators over bool

    def __invert__(self):
        return Not(self)


class All(CompilableSQL):

    def as_sql(self):
        return '*'

    def get_params(self):
        return []


#### Functions

class Function(SQLExpression):
    """
    Represents any mathematical function x, y, ... -> f(x, y, ...)
    It has arguments and returns a value.
    """
    _function = ''

    # number of arguments; if None use arbitrary (len(self._value))
    _arguments_num = 1

    def __init__(self, values):
        if self._arguments_num is not None and\
           len(values) != self._arguments_num:
            raise IndexError('len of argument of "%s" must be %d' %
                             (self.__class__.__name__, self._arguments_num))
        super(Function, self).__init__(values)

    def type(self):
        """
        By default, the type is of the the first
        argument.
        """
        return self._value[0].type()

    @staticmethod
    def _format_parameters(iterable):
        return ', '.join([value.as_sql() for value in iterable])

    def as_sql(self):
        return '{function}({parameters_sql})'\
            .format(function=self._function,
                    parameters_sql=self._format_parameters(self._value))

    def get_params(self):
        params = []
        for value in self._value:
            params += value.get_params()
        return params

    def resolve_columns(self, index):
        self._value = [value.resolve_columns(index) for value in self._value]
        return super(Function, self).resolve_columns(index)


class UnitaryFunction(Function):
    """
    A helper for 1 argument functions.
    """
    _arguments_num = 1

    def __init__(self, argument):
        super(UnitaryFunction, self).__init__([argument])


class Not(UnitaryFunction):
    def as_sql(self):
        return 'NOT (%s)' % self._value[0].as_sql()


class Neg(UnitaryFunction):
    def as_sql(self):
        return '-%s' % self._value[0].as_sql()


class Match(UnitaryFunction):
    _function = 'MATCH'

    def __init__(self, argument):
        #Match always receives a string
        assert isinstance(argument, str)

        # See issue #6:
        # slash `/` must be escaped on MATCH.
        argument = argument.replace('/', r'\/')

        super(UnitaryFunction, self).__init__([String(argument)])


class Count(UnitaryFunction):
    _function = 'COUNT'


class Power(Function):
    _function = 'POW'
    _arguments_num = 2

    def __init__(self, base, exponent):
        super(Power, self).__init__([base, exponent])


class In(Function):
    _function = 'IN'
    _arguments_num = None

    def type(self):
        return Bool

    def __init__(self, lhs, rhs):
        assert isinstance(rhs, (tuple, set, list))
        values = (lhs,) + tuple(convert(value) for value in rhs)
        super(In, self).__init__(values)

    def as_sql(self):
        return '{lhs} {function} ({rhs})'.format(
            function=self._function,
            lhs=self._value[0].as_sql(),
            rhs=self._format_parameters(self._value[1:]))


class NotIn(In):
    _function = 'NOT IN'


class Between(Function):
    _function = 'BETWEEN'
    _arguments_num = 3

    def type(self):
        return Bool

    def __init__(self, lhs, rhs):
        assert isinstance(rhs, (tuple, set, list))
        assert len(rhs) == 2
        values = (lhs,) + tuple(convert(value) for value in rhs)
        super(Between, self).__init__(values)

    def as_sql(self):
        return '{lhs} {function} {first} AND {second}'.format(
            function=self._function,
            lhs=self._value[0].as_sql(),
            first=self._value[1].as_sql(),
            second=self._value[2].as_sql())


class NotBetween(Between):
    _function = 'NOT BETWEEN'


#### Binary Functions

class BinaryFunction(Function):
    """
    Any function of two arguments that is represented by `<lhs> <operation> <rhs>`
    (e.g. 2 + 3).
    """
    _arguments_num = 2

    def __init__(self, lhs, rhs):
        super(BinaryFunction, self).__init__([lhs, rhs])

    def as_sql(self):
        return '{lhs} {function} {rhs}'.format(function=self._function,
                                               lhs=self._value[0].as_sql(),
                                               rhs=self._value[1].as_sql())


class Add(BinaryFunction):
    _function = '+'


class Subtract(BinaryFunction):
    _function = '-'


class Multiply(BinaryFunction):
    _function = '*'


class Divide(BinaryFunction):
    _function = '/'


#### Boolean Functions

class BooleanOperation(BinaryFunction):
    """
    Represents a binary operation over booleans
    """
    def type(self):
        return Bool


class Equal(BooleanOperation):
    _function = '='


class NotEqual(BooleanOperation):
    _function = '!='


class And(BooleanOperation):
    _function = 'AND'

    def as_sql(self):
        return '({lhs}) {function} ({rhs})'.format(function=self._function,
                                                   lhs=self._value[0].as_sql(),
                                                   rhs=self._value[1].as_sql())


class Or(BooleanOperation):
    _function = 'OR'

    def as_sql(self):
        return '({lhs}) {function} ({rhs})'.format(function=self._function,
                                                   lhs=self._value[0].as_sql(),
                                                   rhs=self._value[1].as_sql())


class GreaterThan(BooleanOperation):
    _function = '>'


class LessThan(BooleanOperation):
    _function = '<'


class LessEqualThan(BooleanOperation):
    _function = '<='


class GreaterEqualThan(BooleanOperation):
    _function = '>='


#### Values

class Value(SQLExpression):
    """
    Represents a constant value.
    """
    _input_python_types = ()
    _python_type = None

    def __init__(self, value):
        if not isinstance(value, self._input_python_types):
            raise TypeError("%s only accepts types %s" %
                            (self.__class__, self._input_python_types))
        if not isinstance(value, self._python_type):
            value = self._python_type(value)
        super(Value, self).__init__(value)

    def type(self):
        return type(self)

    def get_params(self):
        return []

    @staticmethod
    def to_python(db_value):
        return db_value


class Integer(Value):
    _input_python_types = (int, float, bool)
    _python_type = int

    def as_sql(self):
        return '%d' % self._value


class Float(Value):
    _input_python_types = (int, float, bool)
    _python_type = float

    def as_sql(self):
        return '%f' % self._value


class Bool(Value):
    _input_python_types = (int, float, bool)
    _python_type = bool

    def as_sql(self):
        if self._value:
            return 'TRUE'
        else:
            return 'FALSE'


class String(Value):
    _input_python_types = (int, float, bool, str)
    _python_type = str

    def as_sql(self):
        return '%s'

    def get_params(self):
        return [self._value]


class Date(Value):
    _input_python_types = (datetime.date,)
    _python_type = datetime.date

    def as_sql(self):
        return calendar.timegm(self._value.timetuple())

    @staticmethod
    def to_python(db_value):
        dt = datetime.datetime.utcfromtimestamp(db_value)
        if settings.USE_TZ:
            dt = dt.replace(tzinfo=get_current_timezone())
        return dt.date()


class DateTime(Date):
    _input_python_types = (datetime.date, datetime.datetime)
    _python_type = datetime.datetime

    @staticmethod
    def to_python(db_value):
        dt = datetime.datetime.utcfromtimestamp(db_value)
        if settings.USE_TZ:
            dt = dt.replace(tzinfo=get_current_timezone())
        return dt



import datetime

from sphinxql.core.query import Query
from sphinxql.sql import Match
from sphinxql.types import String

from .indexes import AuthorIndex, Author2Index
from .models import Author

from tests import SphinxQLTestCase


class QueryTestCase(SphinxQLTestCase):

    def setUp(self):
        super(QueryTestCase, self).setUp()

        Author.objects.create(first_name='foo', last_name='bar', number=1,
                              time=datetime.datetime(2014, 2, 2, 12, 12, 12))
        self.index()

        self.query = Query()
        self.query.fromm.append(AuthorIndex)

    def test_basic(self):
        column = AuthorIndex.first_name

        self.query.select.append(column, 'name')

        self.assertEqual(self.query.sql(), 'SELECT `id`, `first_name` AS name FROM query_authorindex')

    def test_two_selects(self):
        column = AuthorIndex.number

        self.query.select.append(column, column.name)
        self.query.select.append(column*2, 'ss')

        self.assertEqual(self.query.sql(), 'SELECT `id`, `number` AS number, `number` * 2 AS ss '
                                           'FROM query_authorindex')

    def test_two_froms(self):
        self.query.fromm.append(Author2Index)

        self.assertEqual(self.query.sql(), 'SELECT * '
                                           'FROM query_authorindex, query_author2index')

    def test_with_where(self):
        column = AuthorIndex.number

        self.query.select.append(column, column.name)
        self.query.where = column*2 > 2

        self.assertEqual(self.query.sql(), 'SELECT `id`, `number` AS number FROM query_authorindex '
                                           'WHERE `number` * 2 > 2')

    def test_sql_order_by(self):
        column = AuthorIndex.number

        self.query.order_by.append(column, ascending=True)

        self.assertEqual(self.query.sql(), 'SELECT * FROM query_authorindex '
                                           'ORDER BY `number` ASC')

        self.query.order_by.append(column, ascending=False)

        self.assertEqual(self.query.sql(), 'SELECT * FROM query_authorindex '
                                           'ORDER BY `number` DESC')

        self.query.order_by.append(AuthorIndex.time, ascending=True)

        self.assertEqual(self.query.sql(), 'SELECT * FROM query_authorindex '
                                           'ORDER BY `number` DESC, `time` ASC')

    def test_all(self):
        self.assertEqual(self.query.sql(), 'SELECT * FROM query_authorindex')

    def test_execute(self):
        self.assertEqual(len(self.query), 1)

    def test_results(self):
        results = list(self.query)
        self.assertEqual(len(results), 1)
        results = results[0]

        self.assertEqual(results[1], 'foo')
        self.assertEqual(results[2], 'bar')
        self.assertEqual(results[3], 1)

    def test_limit(self):
        results = list(self.query)
        self.assertEqual(len(results), 1)

        self.query.limit = (0, 10)
        results = list(self.query)
        self.assertEqual(len(results), 1)

        self.query.limit = (2, 10)
        results = list(self.query)
        self.assertEqual(len(results), 0)

    def test_match(self):
        self.query.where = Match("foo")
        self.assertEqual(len(self.query), 1)



import shutil
import os

from io import StringIO
from django.conf import settings
from django.core.management import call_command
from django.test import TransactionTestCase

from sphinxql import configuration

try:
    import pymysql
    pymysql.install_as_MySQLdb()
except ImportError:
    pass


class SphinxQLTestCase(TransactionTestCase):

    def setUp(self):
        super(SphinxQLTestCase, self).setUp()

        self._old_path = settings.INDEXES['path']
        settings.INDEXES['path'] += '_test'

        # Django does not support apps using its connections on loading, see
        # https://docs.djangoproject.com/en/1.7/ref/applications/#django.apps.AppConfig.ready
        # Doing so picks the wrong database for tests.
        # We reconfigure after loading to get the right test database.
        configuration.indexes_configurator.configure()
        configuration.indexes_configurator.output()

        # clean path if it exists from previous tests
        if os.path.exists(settings.INDEXES['path']):
            shutil.rmtree(settings.INDEXES['path'])
        os.mkdir(settings.INDEXES['path'])

        call_command('index_sphinx', update=False, stdout=StringIO())
        call_command('start_sphinx', stdout=StringIO())

    def index(self):
        """
        Used to reindex the sphinx index.
        """
        call_command('index_sphinx', stdout=StringIO())

    def tearDown(self):
        call_command('stop_sphinx', stdout=StringIO())
        shutil.rmtree(settings.INDEXES['path'])
        settings.INDEXES['path'] = self._old_path
        super(SphinxQLTestCase, self).tearDown()



import datetime
from unittest import expectedFailure

from sphinxql.core.base import Or
from sphinxql.exceptions import NotSupportedError
from sphinxql.query import QuerySet
from sphinxql.sql import C, Between

from .indexes import DocumentIndex
from .models import Document

from tests import SphinxQLTestCase


def ids_set(query):
    return {entry.id for entry in query}


class SimpleTestCase(SphinxQLTestCase):

    def setUp(self):
        super(SimpleTestCase, self).setUp()

        self.document = Document.objects.create(
            summary="This is a summary", text="What a nice text",
            date=datetime.date(2015, 2, 2),
            added_time=datetime.datetime(2015, 4, 4, 12, 12, 12),
            number=2)

        # retrieve data as it is stored in DB
        self.document = Document.objects.get(id=self.document.id)

        self.index()

        self.query = QuerySet(DocumentIndex)


class SimpleQuerySetTestCase(SimpleTestCase):
    """
    Tests basic operations of QuerySet
    """
    def test_lookup_fail(self):
        self.assertRaises(KeyError, self.query.filter, C('numberERROR') >= 2)

    def test_filter(self):
        self.assertEqual(len(self.query), 1)

        q = self.query.filter(C('number') >= 2)
        self.assertEqual(len(q), 1)

        q = self.query.filter(C('number') < 2)
        self.assertEqual(len(q), 0)

    def test_filter_between_integer(self):

        self.query = self.query.filter(C('number') |Between| (1, 3))
        self.assertEqual(len(self.query), 1)

        self.query = self.query.filter(C('number') |Between| (10, 20))
        self.assertEqual(len(self.query), 0)

    def test_filter_between_dates(self):
        self.query = self.query.filter(C('date') |Between| (datetime.date(2014, 1, 1),
                                                            datetime.date(2016, 1, 1)))
        self.assertEqual(len(self.query), 1)

        self.query = self.query.filter(C('date') |Between| (datetime.date(2000, 1, 1),
                                                            datetime.date(2001, 1, 1)))
        self.assertEqual(len(self.query), 0)

    def test_filter_between_datetimes(self):
        self.query = self.query.filter(C('added_time') |Between| (datetime.datetime(2014, 1, 1),
                                                                  datetime.datetime(2016, 1, 1)))
        self.assertEqual(len(self.query), 1)

        self.query = self.query.filter(C('added_time') |Between| (datetime.datetime(2000, 1, 1),
                                                                  datetime.datetime(2001, 1, 1)))
        self.assertEqual(len(self.query), 0)

    def test_two_filters(self):
        q = self.query.filter(C('number') >= 2)
        q = q.filter(C('date') == datetime.date(2015, 2, 2))
        self.assertEqual(len(q), 1)

    @expectedFailure
    def test_exclude(self):
        """
        SphinxQL does not accept queries of the form "WHERE NOT condition",
        see http://sphinxsearch.com/bugs/view.php?id=2004.
        """
        q = self.query.filter(~(C('number') == 2))
        self.assertEqual(len(q), 0)

    def test_filter_string(self):
        q = self.query.filter(C('summary') == 'This is a summary')
        self.assertEqual(len(q), 1)

        q = self.query.filter(C('summary') == 'This')
        self.assertEqual(len(q), 0)

    def test_search(self):
        q = self.query.search('@text NICE')
        self.assertEqual(len(q), 1)

        q = self.query.search('@text DDDDFDFD')
        self.assertEqual(len(q), 0)

    @expectedFailure
    def test_or(self):
        """
        OR not defined in SphinxQL.
        """
        q = self.query.filter(Or(C('number') == 2, C('number') > 2))
        self.assertEqual(len(q), 1)

    @expectedFailure
    def test_compare_columns(self):
        """
        SphinxQL does not allow comparison between columns,
        see http://sphinxsearch.com/bugs/view.php?id=2015.
        """
        q = self.query.filter(C('added_time') > C('date'))
        self.assertEqual(len(q), 1)

    def test_date_comparison(self):
        q = self.query.filter(C('date') < datetime.datetime(2016, 1, 1))
        self.assertEqual(len(q), 1)

    def test_main_sql(self):
        sql = self.query.query.as_sql()
        self.assertEqual(sql, 'SELECT `id`, `summary`, `text`, `date`, `added_time`, `number` '
                              'FROM queryset_documentindex')

    def test_check_fields(self):
        result = self.query[0]

        self.assertEqual(result.summary, self.document.summary)
        self.assertEqual(result.text, self.document.text)
        self.assertEqual(result.date, self.document.date)
        self.assertEqual(result.added_time, self.document.added_time)
        self.assertEqual(result.number, self.document.number)


class QuerySetLookupTestCase(SimpleTestCase):

    def test_only_1_lookup(self):
        with self.assertRaises(NotImplementedError):
            self.query.filter(string__upper__gte=2)

    def test_invalid_lookup(self):
        with self.assertRaises(KeyError):
            self.query.filter(number__gteERROR=2)

    def test_filter(self):
        self.assertEqual(len(self.query), 1)

        q = self.query.filter(number__gte=2)
        self.assertEqual(len(q), 1)

        q = self.query.filter(number__lt=2)
        self.assertEqual(len(q), 0)

    def test_single_column(self):
        q = self.query.filter(number=2)
        self.assertEqual(len(q), 1)

    def test_tuple_on_rhs(self):
        q = self.query.filter(number__in=(2, 3))
        self.assertEqual(len(q), 1)

    def test_id_in_lhs(self):
        q = self.query.filter(id__in=(1, 2))

    def test_order_by_lookup(self):
        with self.assertRaises(NotImplementedError):
            self.query.order_by('number__max')

    def test_order_by(self):
        q = self.query.order_by('number')
        self.assertEqual(len(q), 1)

        q = self.query.order_by('-number')
        self.assertEqual(len(q), 1)


class QuerySetTestCase(SphinxQLTestCase):

    def setUp(self):
        super(QuerySetTestCase, self).setUp()

        for x in range(1, 101):
            Document.objects.create(
                summary="This is a summary", text="What a nice text. "*x,
                date=datetime.date(2015, 2, 2) + datetime.timedelta(days=x),
                added_time=datetime.datetime(2015, 4, 4, 12, 12, 12),
                number=x*2)

        self.index()

    def test_len(self):
        self.assertEqual(len(QuerySet(DocumentIndex)), 100)
        self.assertEqual(QuerySet(DocumentIndex).count(), 100)

        self.assertEqual(len(QuerySet(DocumentIndex).search('@text adasdsa')), 0)
        self.assertEqual(QuerySet(DocumentIndex).search('@text adasdsa').count(), 0)

    def test_getitem(self):
        query = QuerySet(DocumentIndex)

        self.assertEqual(len(query), 100)
        self.assertEqual(len(query[:20]), 20)

        self.assertEqual(len(query[90:110]), 10)

        self.assertRaises(NotSupportedError, query.__getitem__, slice(90, None))

        self.assertEqual(query[0].number, 2)

    def test_order_by(self):
        q = QuerySet(DocumentIndex).order_by(C('number'))
        self.assertEqual(q[0].number, 2)

        q = QuerySet(DocumentIndex).order_by(-C('number'))
        self.assertEqual(q[0].number, 200)

        q = QuerySet(DocumentIndex).search('@text What').order_by(C('@relevance'))
        # most relevance is last, because has the most occurrences.
        self.assertEqual(q[0].number, 200)

        # other ordering
        q = QuerySet(DocumentIndex).search('@text What').order_by(C('number'))
        self.assertEqual(q[0].number, 2)


class LargeQuerySetTestCase(SphinxQLTestCase):

    def setUp(self):
        super(LargeQuerySetTestCase, self).setUp()

        for x in range(1, 1005):
            Document.objects.create(
                summary="This is a summary", text="What a nice text. "*x,
                date=datetime.date(2015, 2, 2) + datetime.timedelta(days=x),
                added_time=datetime.datetime(2015, 4, 4, 12, 12, 12),
                number=x*2)

        self.documents = Document.objects.all()

        self.index()

    def test_count(self):
        self.assertEqual(QuerySet(DocumentIndex).count(),
                         self.documents.count())

    def test_iterate(self):
        with self.assertRaises(IndexError):
            iter(QuerySet(DocumentIndex))

        self.assertEqual(len(list(iter(QuerySet(DocumentIndex)[:1000]))), 1000)

    def test_filter(self):
        sphinx_query = QuerySet(DocumentIndex).filter(number__gt=30)[:1000]
        django_query = self.documents.filter(number__gt=30)

        self.assertEqual(ids_set(sphinx_query), ids_set(django_query))



from ..core.base import SQLExpression, Integer


def quote(string):
    return "`%s`" % string


class Column(SQLExpression):
    """
    Represents a column in the database.

    _value is the alias of the column, _type is the type
    it represents.
    """
    def __init__(self, type, alias):
        super(Column, self).__init__(alias)
        self._type = type

    def as_sql(self):
        return quote('%s' % self._value)

    def type(self):
        return self._type

    @property
    def name(self):
        return self._value

    def get_params(self):
        return []


class IdColumn(Column):
    """
    Column representing the document id
    """
    def __init__(self):
        super(IdColumn, self).__init__(Integer, 'id')


class WeightColumn(Column):
    """
    Column representing the document id
    """
    def __init__(self):
        super(WeightColumn, self).__init__(Integer, 'weight()')

    def as_sql(self):
        return '%s' % self._value



from collections import OrderedDict
import os.path
import re

from django.conf import settings
from django.db import connections
from django.db.models import F
from django.db.models.expressions import Combinable

from ..exceptions import ImproperlyConfigured
from ..types import DateTime, Date
from .configurations import IndexerConfiguration, \
    SearchdConfiguration, \
    SourceConfiguration, \
    IndexConfiguration
from . import constants


def add_source_conf_param(source_conf, key, value):
    """
    Adds a {key: value} to source_conf taking into account if
    key is a multi_valued_parameter or single_valued_parameter.
    """
    if key in constants.source_multi_valued_parameters:
        if key not in source_conf:
            source_conf[key] = []
        source_conf[key].append(value)
    elif key in constants.source_single_valued_parameters:
        source_conf[key] = value
    else:
        raise ImproperlyConfigured('Invalid parameter "%s" in source '
                                   'configuration' % key)
    return source_conf


SPHINX_TO_DJANGO_MAP = {'user': 'sql_user',
                        'passwd': 'sql_pass', 'password': 'sql_pass',
                        'db': 'sql_db', 'database': 'sql_db',
                        'port': 'sql_port',
                        'host': 'sql_host',
                        'unix_socket': 'sql_sock'
                        }

DJANGO_TO_SPHINX_VENDOR = {'postgresql': 'pgsql',
                           'mysql': 'mysql'}

DEFAULT_INDEXER_PARAMS = {}

DEFAULT_SOURCE_PARAMS = {'sql_host': 'localhost',
                         'sql_pass': '',
                         }

DEFAULT_SEARCHD_PARAMS = {'listen': '9306:mysql41',
                          }

DEFAULT_INDEX_PARAMS = {'type': 'plain'}


def _pymysql_mogrify(cursor, query, args=None):
        """
        This is a copy of the code that is executed in PyMySQL Cursor.execute
        function, but only builds the string; it does not execute it.
        """
        conn = cursor._get_db()

        if args is not None:
            query = query % cursor._escape_args(args, conn)

        return query


def _generate_sql(query, vendor):
    """
    Returns a string with the SQL generated by the particular vendor.

    vendor can only be 'pgsql' or 'mysql'.
    """
    assert(vendor in ('mysql', 'pgsql'))

    db = query.db
    compiler = query.query.get_compiler(using=db)
    cursor = compiler.connection.cursor()
    if vendor == 'pgsql':
        return cursor.mogrify(*compiler.as_sql()).decode('utf-8')
    else:
        if hasattr(cursor, 'mogrify'):
            # the backend has `mogrify` to map sql, params -> sql
            return cursor.mogrify(*compiler.as_sql())
        else:
            # backend doesn't have. We use the code from PyMySQL
            return _pymysql_mogrify(cursor, *compiler.as_sql())


def _build_query(index, query, vendor):
    """
    Returns a SQL query built according to the fields we want to index.
    """
    assert(vendor in ('mysql', 'pgsql'))

    def special_annotate(query, dict_values):
        """
        This is a copy of normal Django annotate but 1. does not check for name
        collisions and 2. does not add GROUP_BY to the query.

        GROUP BY is not required since we are building a SELECT expression.
        Name collision is avoided so our indexes can have the same names as
        the Model.
        """
        obj = query._clone()

        # this line de-activates GROUP BY so we remember what magic we did it.
        # obj._setup_aggregate_query(list(dict_values.items()))

        # Add the aggregates to the query
        for (alias, aggregate_expr) in dict_values.items():
            obj.query.add_annotation(aggregate_expr, alias, is_summary=False)
            # obj.query.add_aggregate(aggregate_expr, query.model, alias,
            #                         is_summary=False)

        return obj

    annotation = OrderedDict()
    for field in index.Meta.fields:
        if isinstance(field.model_attr, str):
            annotation[field.name] = F(field.model_attr)
        elif isinstance(field.model_attr, Combinable):
            annotation[field.name] = field.model_attr
        else:
            raise ImproperlyConfigured(
                'Field "%s" model_attr must be either '
                'a string or a F expression. It is a "%s".' %
                (field.name, type(field.model_attr)))

    # this is an hacky approach, but until we find something better,
    # we have to live with it.
    query = special_annotate(query.only('id'), annotation)
    sql = _generate_sql(query, vendor)

    for field in [f for f in index.Meta.fields
                  if f.type() in (DateTime, Date)]:
        alias = '`%s`' % field.name
        expression = '%s'
        if vendor == 'mysql':
            expression = "UNIX_TIMESTAMP(CONVERT_TZ(" \
                         "%s, " \
                         "'+00:00', " \
                         "@@session.time_zone))"

        elif vendor == 'pgsql':
            alias = '"%s"' % field.name
            if field.type() is DateTime:
                expression = "EXTRACT(EPOCH FROM %s AT TIME ZONE '%s')" % \
                             ('%s', settings.TIME_ZONE)
            elif field.type() is Date:
                expression = "EXTRACT(EPOCH FROM %s)"

        sql = re.sub('([^\s]*) AS %s' % alias,
                     lambda m: '%s AS %s' % (expression % m.group(1),
                                             alias), sql)
    return sql


class Configurator(object):
    """
    The main configurator.

    Uses the settings dictionary ``INDEXES``.
    """
    def __init__(self):
        if not hasattr(settings, 'INDEXES'):
            raise ImproperlyConfigured('Django-SphinxQL requires '
                                       'settings.INDEXES')

        sphinx_path = settings.INDEXES.get('sphinx_path')
        self.sphinx_file = os.path.join(sphinx_path, 'sphinx.conf')

        # registered indexes
        self._registered_indexes = []

        # configured indexes
        self.indexes = OrderedDict()
        self.indexes_confs = []
        self.sources_confs = []

        # configured indexer and searchd
        self.indexer_conf = None
        self.searchd_conf = None

    def register(self, index):
        """
        Registers an index into the configuration.

        Used by index to register itself in the configuration.
        """
        if index not in self._registered_indexes:
            self._registered_indexes.append(index)

    @staticmethod
    def _configure_source(index):
        """
        Maps an ``Index`` into a Sphinx source configuration.
        """
        source_attrs = OrderedDict()
        source_attrs.update(DEFAULT_SOURCE_PARAMS)
        source_attrs.update(settings.INDEXES.get('source_params', {}))
        source_attrs.update(getattr(index.Meta, 'source_params', {}))

        if hasattr(index.Meta, 'query'):
            query = index.Meta.query
        else:
            query = index.Meta.model.objects.all()

        ### select type from backend
        if connections[query.db].vendor not in DJANGO_TO_SPHINX_VENDOR:
            raise ImproperlyConfigured('Django-SphinxQL currently only supports '
                                       'mysql and postgresql backends')
        vendor = DJANGO_TO_SPHINX_VENDOR[connections[query.db].vendor]

        source_attrs = add_source_conf_param(source_attrs, 'type', vendor)

        if vendor == 'mysql':
            source_attrs = add_source_conf_param(source_attrs, 'sql_query_pre',
                                                 'SET CHARACTER_SET_RESULTS=utf8')

        ### build connection parameters from Django connection parameters
        connection_params = connections[query.db].get_connection_params()
        for key in connection_params:
            if key in SPHINX_TO_DJANGO_MAP:
                value = SPHINX_TO_DJANGO_MAP[key]
                source_attrs = add_source_conf_param(source_attrs,
                                                     value,
                                                     connection_params[key])

        ### create parameters for attributes
        for field in index.Meta.fields:
            if field.is_attribute:
                source_attrs = add_source_conf_param(source_attrs,
                                                     field._sphinx_field_name,
                                                     field.name)

        if hasattr(index.Meta, 'range_step'):
            # see http://sphinxsearch.com/docs/current.html#ranged-queries
            range_step = int(index.Meta.range_step)

            source_attrs = add_source_conf_param(
                source_attrs, 'sql_query_range',
                'SELECT MIN(id),MAX(id) FROM %s' % index.Meta.model._meta.db_table)
            source_attrs = add_source_conf_param(
                source_attrs, 'sql_range_step', range_step)
            query = query.extra(where=['{0}.id>=$start AND {0}.id<=$end'
                                .format(index.Meta.model._meta.db_table)])

        ### add the query
        source_attrs = add_source_conf_param(
            source_attrs,
            'sql_query', _build_query(index, query, vendor))

        return SourceConfiguration(index.build_name(), source_attrs)

    @staticmethod
    def _configure_index(index, source_name):
        """
        Maps a ``Index`` into a Sphinx index configuration.
        """
        index_params = {
            'source': source_name,
            'path': os.path.join(settings.INDEXES['path'], source_name)
        }
        index_params.update(DEFAULT_INDEX_PARAMS)
        index_params.update(settings.INDEXES.get('index_params', {}))
        index_params.update(getattr(index.Meta, 'index_params', {}))

        return IndexConfiguration(index.build_name(), index_params)

    @staticmethod
    def _configure_searchd():
        searchd_params = OrderedDict()
        searchd_params.update(DEFAULT_SEARCHD_PARAMS)
        searchd_params['pid_file'] = os.path.join(settings.INDEXES.get('sphinx_path'), 'searchd.pid')
        # see WARNING at http://sphinxsearch.com/docs/current/conf-binlog-path.html
        searchd_params['binlog_path'] = settings.INDEXES.get('sphinx_path')

        searchd_params.update(settings.INDEXES.get('searchd_params', {}))

        return SearchdConfiguration(searchd_params)

    @staticmethod
    def _configure_indexer():
        indexer_params = OrderedDict()
        indexer_params.update(DEFAULT_INDEXER_PARAMS)
        indexer_params.update(settings.INDEXES.get('indexer_params', {}))
        return IndexerConfiguration(indexer_params)

    def configure(self):
        """
        Configures the registered indexes.

        This method must be called before `output`.
        """
        if not hasattr(settings, 'INDEXES'):
            raise ImproperlyConfigured('Django-SphinxQL requires '
                                       'settings.INDEXES')

        self.indexer_conf = self._configure_indexer()
        self.searchd_conf = self._configure_searchd()
        self.sources_confs.clear()
        self.indexes_confs.clear()
        self.indexes.clear()
        for index in self._registered_indexes:
            meta = getattr(index.Meta.model, '_meta', None)
            assert meta is not None

            assert index not in self.indexes.values()
            self.indexes[index.build_name()] = index

            source_conf = self._configure_source(index)
            index_conf = self._configure_index(index, source_conf.name)
            self.sources_confs.append(source_conf)
            self.indexes_confs.append(index_conf)

    def output(self):
        """
        Outputs the configuration file `sphinx.conf`.
        """
        assert self.indexes

        string_blocks = ["# WARNING! This file was automatically generated: do not "
                         "modify it.\n"]
        # output all source and indexes
        for i in range(len(self.indexes)):
            string_blocks.append(self.sources_confs[i].format_output())
            string_blocks.append(self.indexes_confs[i].format_output())

        # output indexer and searchd
        string_blocks.append(self.indexer_conf.format_output())
        string_blocks.append(self.searchd_conf.format_output())

        with open(self.sphinx_file, 'w') as conf_file:
            conf_file.write('\n'.join(string_blocks))



from __future__ import unicode_literals

from django.core.management.base import BaseCommand

from sphinxql import configuration


class Command(BaseCommand):
    help = 'Interacts with search deamon.'

    def handle(self, **options):
        self.stdout.write('Starting Sphinx')
        self.stdout.write('---------------')

        self.stdout.write(configuration.start())

        self.stdout.write('----')
        self.stdout.write('Done')



from django.db import models


class Author(models.Model):
    first_name = models.CharField(max_length=30)
    last_name = models.CharField(max_length=30)
    number = models.IntegerField()

    time = models.DateTimeField()



from sphinxql import indexes, fields

from .models import Author


class AuthorIndex(indexes.Index):
    first_name = fields.IndexedString(model_attr='first_name')
    last_name = fields.IndexedString(model_attr='last_name')
    number = fields.Integer(model_attr='number')
    time = fields.DateTime(model_attr='time')

    class Meta:
        model = Author
        query = Author.objects.all()


class Author2Index(indexes.Index):
    first_name = fields.IndexedString(model_attr='first_name')

    class Meta:
        model = Author
        query = Author.objects.all()



from unittest import TestCase

from sphinxql.sql import Column
from sphinxql.types import Integer
from sphinxql.core.query import SelectStatement, FromStatement


class SelectStatementTestCase(TestCase):
    def test_basic(self):
        select = SelectStatement()

        column = Column(Integer, 'test')

        select.append(column, column.name)

        self.assertEqual(select.as_sql(), '`id`, `test` AS test')

    def test_two_columns(self):
        select = SelectStatement()

        column = Column(Integer, 'test')

        select.append(column, column.name)
        select.append(column*2, 'ss')
        select.append(column*3)

        self.assertEqual(select.sql(), '`id`, `test` AS test, `test` * 2 AS ss, `test` * 3')


class MockIndex:

    def __init__(self, name):
        self.name = name

    def build_name(self):
        return self.name


class FromStatementTestCase(TestCase):

    def test_basic(self):
        fromm = FromStatement()

        fromm.append(MockIndex('test'))

        self.assertEqual(fromm.sql(), 'test')

    def test_two_indexes(self):
        fromm = FromStatement()

        fromm.append(MockIndex('test'))
        fromm.append(MockIndex('test1'))

        self.assertEqual(fromm.sql(), 'test, test1')

    def test_raise_at_same_names(self):
        fromm = FromStatement()

        fromm.append(MockIndex('test'))
        with self.assertRaises(AssertionError):
            fromm.append(MockIndex('test'))



from collections import OrderedDict
from copy import deepcopy

from ..configuration.connection import Connection, DEFAULT_LIMIT_COUNT
from ..exceptions import NotSupportedError
from .base import CompilableSQL, All
from .columns import IdColumn, Column

# Sphinx sets a max of 5 columns in order by.
MAX_ORDER_BY_ALLOWED = 5


class Query(CompilableSQL):
    """
    A SphinxQL Query.
    """
    def __init__(self, connection=None):
        self._statements = {
            'select': SelectStatement(),
            'from': FromStatement(),
            'where': None,
            'order_by': OrderByStatement(),
            'limit': None   # `None` or (offset, count)
        }
        self.low_mark, self.high_mark = 0, None

        self._connection = connection
        if connection is None:
            self._connection = Connection()

    def __iter__(self):
        """
        If limits are defined, returns an iterator over all results.
        else, iterates over all results using LIMIT in chunks of
        DEFAULT_LIMIT_COUNT.
        """
        return self._connection.iterator(self.as_sql(), self.get_params())

    def __str__(self):
        return self.as_sql() % tuple("\"%s\"" % x for x in self.get_params())

    def __len__(self):
        return len(list(iter(self)))

    @property
    def select(self):
        return self._statements['select']

    @select.setter
    def select(self, value):
        self._statements['select'] = value

    @property
    def fromm(self):
        return self._statements['from']

    @fromm.setter
    def fromm(self, value):
        self._statements['from'] = value

    @property
    def where(self):
        return self._statements['where']

    @where.setter
    def where(self, value):
        self._statements['where'] = value

    @property
    def order_by(self):
        return self._statements['order_by']

    @property
    def limit(self):
        return self._statements['limit']

    @limit.setter
    def limit(self, value):
        if value is not None:
            assert isinstance(value, tuple)
            assert len(value) == 2 and isinstance(value[0], int) and isinstance(value[1], int)
        self._statements['limit'] = value

    def as_sql(self):
        statements = self._statements.copy()
        if not statements['select']:
            statements['select'] = All()

        query = 'SELECT {select} FROM {from}'

        if statements['where']:
            query += ' WHERE {where}'
        if statements['order_by']:
            query += ' ORDER BY {order_by}'
        if statements['limit']:
            query += ' LIMIT %d, %d' % statements['limit']

        cleaned_statements = {key: value.as_sql() for key, value in
                              statements.items() if value and key != 'limit'}

        return query.format(**cleaned_statements)

    def get_params(self):
        params = []
        for clause in ('select', 'from', 'where'):
            if self._statements[clause]:
                params += self._statements[clause].get_params()
        return params

    def clone(self):
        clone = Query()
        # deep because we want new statements
        clone._statements = deepcopy(self._statements)
        return clone


class SelectStatement(CompilableSQL):

    def __init__(self):
        self._expressions = []
        self._alias = []  # list of alias (string) or None if no alias for expression

    def __len__(self):
        """
        For the existence of query
        """
        return len(self._expressions)

    def clear(self):
        self._expressions.clear()
        self._alias.clear()

    def append(self, expression, alias=None):
        assert (alias is None) or alias not in self._alias

        # first column must be ID
        if not self._expressions:
            self._expressions.append(IdColumn())
            self._alias.append(None)
        self._expressions.append(expression)
        self._alias.append(alias)

    def as_sql(self):
        sql = ''
        first = True
        for pos, alias in enumerate(self._alias):
            if first:
                separator = ''
                first = False
            else:
                separator = ', '

            if alias:
                sql += separator + '%s AS %s' % (self._expressions[pos].as_sql(), alias)
            else:
                sql += separator + self._expressions[pos].as_sql()
        return sql

    def get_params(self):
        params = []
        for x in self._expressions:
            params += x.get_params()
        return params


class FromStatement(CompilableSQL):

    def __init__(self):
        self._indexes = OrderedDict()  # index_name: index

    def append(self, index):
        assert index.build_name() not in self._indexes
        self._indexes[index.build_name()] = index

    def as_sql(self):
        assert self._indexes
        sql = ''

        first = True
        for index_name in self._indexes:
            if first:
                separator = ''
                first = False
            else:
                separator = ', '

            sql += separator + '%s' % index_name

        return sql

    def get_params(self):
        return []


class OrderByStatement(CompilableSQL):
    """
    Used to create order by statements
    """
    _DIRECTION = {True: 'ASC', False: 'DESC'}

    def __init__(self):
        self._columns = []  # columns to as_sql()
        self._directions = []  # 'ASC' or 'DESC'
        self._columns_names = []  # columns names already inside

    def __len__(self):
        """
        For the existence of query
        """
        return len(self._columns)

    def as_sql(self):
        assert self._columns
        sql = ''

        first = True
        for pos, column in enumerate(self._columns):
            if first:
                separator = ''
                first = False
            else:
                separator = ', '

            sql += separator + '%s %s' % (column.as_sql(), self._directions[pos])

        return sql

    def append(self, column, ascending=True):
        assert isinstance(column, Column)
        assert isinstance(ascending, bool)
        if len(self._columns) == MAX_ORDER_BY_ALLOWED:
            raise NotSupportedError("Sphinx only supports up to %s expressions "
                                    "in order by" % MAX_ORDER_BY_ALLOWED)

        direction = self._DIRECTION[ascending]

        if column.name in self._columns_names:
            index = self._columns_names.index(column.name)
            self._columns[index] = column
            self._directions[index] = direction
        else:
            self._columns_names.append(column.name)
            self._columns.append(column)
            self._directions.append(direction)

    def clear(self):
        self._columns.clear()
        self._directions.clear()
        self._columns_names.clear()



from sphinxql.manager import IndexManager
from sphinxql import indexes, fields

from .models import Document


class Manager(IndexManager):

    def get_queryset(self):
        return super(Manager, self).get_queryset().filter(number__in=[2, 4, 6])


class DocumentIndex(indexes.Index):
    summary = fields.String(model_attr='summary')
    text = fields.IndexedString(model_attr='text')

    date = fields.Date(model_attr='date')
    added_time = fields.DateTime(model_attr='added_time')

    number = fields.Integer(model_attr='number')

    other_objects = Manager()
    objects = IndexManager()

    class Meta:
        model = Document




# for testing postgres
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql_psycopg2',
        'NAME': 'sphinx_example',
        'TEST_NAME': 'sphinx_example_test',
        'USER': 'sphinx_example',
        'PASSWORD': 'test'
    },
}

# DATABASES = {
#     'default': {
#         'ENGINE': 'django.db.backends.mysql',
#         'NAME': 'sphinx_example',
#         'TEST_NAME': 'sphinx_example_test',
#         'USER': 'root',
#         },
# }

INSTALLED_APPS = ('tests.query', 'tests.queryset',
                  'tests.indexing', 'tests.foreign_relationships', 'sphinxql')

SECRET_KEY = "django_tests_secret_key"

MIDDLEWARE_CLASSES = ()

# Use a fast hasher to speed up tests.
PASSWORD_HASHERS = (
    'django.contrib.auth.hashers.MD5PasswordHasher',
)

import os
BASE_DIR = os.path.dirname(os.path.dirname(__file__))

# we add 'U+00E2' to test unicode
# we cannot override settings since this is used at config time.
INDEXES = {
    'path': os.path.join(BASE_DIR, '_index'),
    'sphinx_path': BASE_DIR,
    'index_params': {'charset_table': '0..9, A..Z->a..z, _, a..z, /, '
                     'U+00E2'}
}




source_single_valued_parameters = (
    'type',
    'sql_host', 'sql_user', 'sql_pass', 'sql_db', 'sql_port',
    'sql_sock',

    'json_autoconv_numbers',
    'json_autoconv_keynames',
    'on_json_attr_error',

    'mysql_connect_flags',
    'mysql_ssl_cert', 'mysql_ssl_key', 'mysql_ssl_ca',

    'mssql_winauth', 'mssql_unicode',

    'odbc_dsn',

    'sql_column_buffers',

    'sql_query',
    'sql_joined_field',
    'sql_query_post',
    'sql_query_post_index',
    'sql_ranged_throttle',
    'sql_query_range',
    'sql_range_step',
    'sql_query_killlist',
    'sql_query_info',

    'unpack_zlib',
    'unpack_mysqlcompress',
    'unpack_mysqlcompress_maxsize'
)

source_multi_valued_parameters = (
    'sql_query_pre',

    'xmlpipe_attr_uint',
    'xmlpipe_attr_bigint',
    'xmlpipe_attr_bool',
    'xmlpipe_attr_timestamp',
    'xmlpipe_attr_str2ordinal',
    'xmlpipe_attr_float',
    'xmlpipe_attr_multi',
    'xmlpipe_attr_multi_64',
    'xmlpipe_attr_string',
    'xmlpipe_attr_wordcount',
    'xmlpipe_attr_json',
    'xmlpipe_fixup_utf8',

    'xmlpipe_command',
    'xmlpipe_field',
    'xmlpipe_field_string',
    'xmlpipe_field_wordcount',

    'sql_attr_uint',
    'sql_attr_timestamp',
    'sql_attr_float',
    'sql_attr_str2ordinal',
    'sql_attr_string',
    'sql_attr_json',
    'sql_attr_multi',
    'sql_attr_bool',
    'sql_attr_bigint',
    'sql_attr_str2wordcount',

    'sql_field_string',
    'sql_field_str2wordcount',
    'sql_file_field'
)

source_parameters = source_single_valued_parameters + source_multi_valued_parameters

source_mandatory_parameters = (
    'type',
    'sql_host',
    'sql_user',
    'sql_pass',
    'sql_db',
    'sql_query',
)

source_limited_options = {
    'type': ('mysql', 'pgsql', 'mssql', 'xmlpipe', 'xmlpipe2', 'odbc'),
}


index_single_valued_parameters = (
    'type',
    'source',
    'path',
    'docinfo',
    'mlock',
    'morphology',
    'dict',
    'index_sp',
    'index_zones',
    'min_stemming_len',
    'stopwords',
    'wordforms',
    'embedded_limit',
    'exceptions',
    'min_word_len',
    'charset_type',
    'charset_table',
    'ignore_chars',
    'min_prefix_len',
    'min_infix_len',
    'max_substring_len',
    'prefix_fields',
    'infix_fields',
    'enable_star',
    'ngram_len',
    'ngram_chars',
    'phrase_boundary',
    'phrase_boundary_step',
    'html_strip',
    'html_index_attrs',
    'html_remove_elements',
    'local',
    'agent',
    'agent_persistent',
    'agent_blackhole',
    'agent_connect_timeout',
    'agent_query_timeout',
    'preopen',
    'ondisk_dict',
    'inplace_enable',
    'inplace_hit_gap',
    'inplace_docinfo_gap',
    'inplace_reloc_factor',
    'inplace_write_factor',
    'index_exact_words',
    'overshort_step',
    'stopword_step',
    'hitless_words',
    'expand_keywords',
    'blend_chars',
    'blend_mode',
    'rt_mem_limit',
    'ha_strategy',
    'bigram_freq_words',
    'bigram_index',
    'index_field_lengths',
    'regexp_filter',
    'stopwords_unstemmed',
    'global_idf',
)

index_multi_valued_parameters = (
    'rt_field',
    'rt_attr_uint',
    'rt_attr_bool',
    'rt_attr_bigint',
    'rt_attr_float',
    'rt_attr_multi',
    'rt_attr_multi_64',
    'rt_attr_timestamp',
    'rt_attr_string',
    'rt_attr_json',
)

index_parameters = index_single_valued_parameters + index_multi_valued_parameters

index_mandatory_parameters = (
    'source',
    'path',
)

indexer_parameters = (
    'mem_limit',
    'max_iops',
    'max_iosize',
    'max_xmlpipe2_field',
    'write_buffer',
    'max_file_field_buffer',
    'on_file_field_error',
    'lemmatizer_base',
    'lemmatizer_cache',
)

searchd_parameters = (
    'listen',
    'address',
    'log',
    'query_log',
    'query_log_format',
    'read_timeout',
    'client_timeout',
    'max_children',
    'pid_file',
    'max_matches',
    'seamless_rotate',
    'preopen_indexes',
    'unlink_old',
    'attr_flush_period',
    'ondisk_dict_default',
    'max_packet_size',
    'mva_updates_pool',
    'crash_log_path',
    'max_filters',
    'max_filter_values',
    'listen_backlog',
    'read_buffer',
    'read_unhinted',
    'max_batch_queries',
    'subtree_docs_cache',
    'subtree_hits_cache',
    'workers',
    'dist_threads',
    'binlog_path',
    'binlog_flush',
    'binlog_max_log_size',
    'snippets_file_prefix',
    'collation_server',
    'collation_libc_locale',
    'plugin_dir',
    'mysql_version_string',
    'rt_flush_period',
    'thread_stack',
    'expansion_limit',
    'compat_sphinxql_magics',
    'watchdog',
    'prefork_rotation_throttle',
    'sphinxql_state',
    'ha_ping_interval',
    'ha_period_karma',
    'persistent_connections_limit',
    'rt_merge_iops',
    'rt_merge_maxiosize',
    'predicted_time_costs',
)

searchd_mandatory_parameters = (
    'pid_file',
)

reserved_keywords = (
    'AND',
    'AGENT',
    'AS',
    'ASC',
    'AVG',
    'BEGIN',
    'BETWEEN',
    'BY',
    'CALL',
    'COLLATION',
    'COMMIT',
    'COUNT',
    'DELETE',
    'DESC',
    'DESCRIBE',
    'DISTINCT',
    'FALSE',
    'FROM',
    'GLOBAL',
    'GROUP',
    'ID',
    'IN',
    'INSERT',
    'INTO',
    'LIMIT',
    'MATCH',
    'MAX',
    'META',
    'MIN',
    'NOT',
    'NULL',
    'OPTION',
    'OR',
    'ORDER',
    'REPLACE',
    'ROLLBACK',
    'SELECT',
    'SET',
    'SHOW',
    'START',
    'STATUS',
    'SUM',
    'TABLES',
    'TRANSACTION',
    'TRUE',
    'UPDATE',
    'VALUES',
    'VARIABLES',
    'WARNINGS',
    'WEIGHT',
    'WHERE',
    'WITHIN',
)



import datetime
from unittest import TestCase

from sphinxql.types import Date, DateTime, Float, String


class DateTestCase(TestCase):

    def test_date(self):
        self.assertRaises(TypeError, Date, 2)
        self.assertRaises(TypeError, Date, 'sql injection')

        try:
            Date(datetime.date(2013, 2, 2))
        except TypeError:
            self.fail("Date() raised TypeError on date.")

        try:
            Date(datetime.datetime.utcnow())
        except TypeError:
            self.fail("Date() raised TypeError on datetime.")

    def test_datetime(self):
        self.assertRaises(TypeError, DateTime, 2)
        self.assertRaises(TypeError, DateTime, 'sql injection')
        self.assertRaises(TypeError, DateTime, datetime.date(2013, 2, 2))

        try:
            DateTime(datetime.datetime.utcnow())
        except TypeError:
            self.fail("DateTime() raised TypeError on datetime.")

    def test_float(self):
        self.assertRaises(TypeError, Float, 'sql injection')

        try:
            Float(True)
        except TypeError:
            self.fail("Float() raised TypeError on bool.")


class ConversionsTestCase(TestCase):

    def test_datetimes(self):

        time = datetime.datetime(2014, 2, 2, 12, 12, 12)

        db_value = int(DateTime(time).as_sql())
        self.assertEqual(DateTime.to_python(db_value), time)

    def test_dates(self):

        date = datetime.date(2014, 2, 2)

        db_value = int(Date(date).as_sql())
        self.assertEqual(Date.to_python(db_value), date)

    def test_float(self):

        f = 3.2

        db_value = float(Float(f).as_sql())
        self.assertEqual(Float.to_python(db_value), f)

    def test_string(self):

        string = 'dsadsa sdsadas sd as'

        db_value = String(string).as_sql() % String(string).get_params()[0]
        self.assertEqual(Float.to_python(db_value), string)



from .core import columns
from . import types


class Field(columns.Column):
    """
    A generic field for indexes.

    attr_name is the attribute it uses from the ORM object.
    """
    _type = None
    _sphinx_field_name = None

    def __init__(self, model_attr):
        super(Field, self).__init__(self._type, None)
        self.model_attr = model_attr

    @property
    def is_attribute(self):
        """
        Returns weather the field is a Sphinx attribute. A Sphinx query only
        returns attributes.
        """
        return self._sphinx_field_name is not None


class Text(Field):
    """
    A Sphinx field (indexed but not stored)
    """
    _sphinx_field_name = None
    _type = types.String


class Integer(Field):
    """
    A Sphinx attribute (i.e. not indexed)
    """
    _sphinx_field_name = 'sql_attr_bigint'
    _type = types.Integer


class Float(Field):
    """
    A Sphinx attribute (i.e. not indexed)
    """
    _sphinx_field_name = 'sql_attr_float'
    _type = types.Float


class String(Field):
    """
    A Sphinx attribute (i.e. not indexed)
    """
    _sphinx_field_name = 'sql_attr_string'
    _type = types.String


class IndexedString(Field):
    """
    A Sphinx field and attribute (i.e. indexed and stored)
    """
    _sphinx_field_name = 'sql_field_string'
    _type = types.String


class DateTime(Field):
    """
    A Sphinx attribute (i.e. not indexed)
    """
    _sphinx_field_name = 'sql_attr_timestamp'
    _type = types.DateTime


class Date(Field):
    _sphinx_field_name = 'sql_attr_timestamp'
    _type = types.Date


class Bool(Field):
    _sphinx_field_name = 'sql_attr_bool'
    _type = types.Bool



# see http://stackoverflow.com/a/21416007/931303
try:
    import pymysql
    pymysql.install_as_MySQLdb()
except ImportError:
    pass

import MySQLdb


DEFAULT_LIMIT_COUNT = 100


class Connection():
    def __init__(self, host=None, port=None):
        self.host, self.port = self.configure_connection(host, port)
        self.db = None

    def iterator(self, sql, params):
        # lazy connect to the server to avoid connection without usage.
        if self.db is None:
            self.db = MySQLdb.connect(host=self.host, port=self.port, charset='utf8')
        cursor = self.db.cursor()

        try:
            cursor.execute(sql, params)
        except Exception:
            cursor.close()
            raise

        cursor.execute(sql, params)

        for x in range(cursor.rowcount):
            yield cursor.fetchone()

        cursor.close()

    @staticmethod
    def configure_connection(host, port):
        from sphinxql.configuration import indexes_configurator
        if host is None:
            host = '0'
        if port is None:
            port_string = indexes_configurator.searchd_conf.params['listen']
            port = int(port_string.split(':')[0])

        return host, port



import logging

from django.apps import AppConfig, apps

from sphinxql.configuration import indexes_configurator
from django.db.utils import InternalError, OperationalError

logger = logging.getLogger(__name__)


class SphinxQL(AppConfig):
    name = 'sphinxql'

    def ready(self):
        """
        Loads all indexes and configures Sphinx. When db isn't ready, it logs
        it and doesn't index the models.
        """
        for app in apps.get_app_configs():
            module_name = app.module.__package__ + '.indexes'
            try:
                __import__(module_name)
            except ImportError as e:
                if module_name not in str(e):
                    raise
                # ignore apps without indexes.
                pass

        try:
            indexes_configurator.configure()
            indexes_configurator.output()
        except (InternalError, OperationalError):
            logger.warning('Sphinx was not configured: no database found.')
            pass



import os

if os.environ['BACKEND'] == 'mysql':
    DATABASES = {
        'default': {
            'ENGINE': 'django.db.backends.mysql',
            'NAME': 'sphinx_example',
            'TEST_NAME': 'sphinx_example_test',
            'USER': 'root',
            },
        }
else:
    DATABASES = {
        'default': {
            'ENGINE': 'django.db.backends.postgresql_psycopg2',
            'NAME': 'sphinx_example',
            'TEST_NAME': 'sphinx_example_test',
            'USER': 'postgres'
            },
        }

INSTALLED_APPS = ('sphinxql', 'tests.query', 'tests.queryset',
                  'tests.indexing', 'tests.foreign_relationships')

SECRET_KEY = "django_tests_secret_key"

MIDDLEWARE_CLASSES = ()

# Use a fast hasher to speed up tests.
PASSWORD_HASHERS = (
    'django.contrib.auth.hashers.MD5PasswordHasher',
)

import os
BASE_DIR = os.path.dirname(os.path.dirname(__file__))

# we add 'U+00E2' to test unicode
# we cannot override settings since this is used at config time.
INDEXES = {
    'path': os.path.join(BASE_DIR, '_index'),
    'sphinx_path': BASE_DIR,
    'index_params': {'charset_table': '0..9, A..Z->a..z, _, a..z, /, '
                                      'U+00E2'},
    'searchd_params': {'binlog_path': ''}
}



import datetime
from sphinxql.core.base import Date

from sphinxql.query import Query

from .indexes import DocumentIndex
from .models import Document, Type, MainType

from tests import SphinxQLTestCase


class IndexTestCase(SphinxQLTestCase):

    def setUp(self):
        super(IndexTestCase, self).setUp()

        main_type = MainType.objects.create(name='MainType1')
        self.date = datetime.datetime.now().date()
        type = Type.objects.create(name='Type1', type=main_type, date=self.date)

        Document.objects.create(type=type, text="What a nice text")
        self.index()

    def test_basic(self):
        query = Query()
        query.fromm.append(DocumentIndex)
        result = list(query)

        self.assertEqual('Type1', result[0][2])
        self.assertEqual('MainType1', result[0][3])
        self.assertEqual('MainType1', result[0][4])
        self.assertEqual('MainType1 Type1', result[0][5])
        self.assertEqual(self.date, Date.to_python(result[0][6]))



from . import constants
from ..exceptions import ImproperlyConfigured


class Configuration(object):
    """
    A generic Sphinx configuration.

    Has a type, a type_name and parameters in a dictionary form.
    """
    type_name = None
    valid_parameters = tuple()
    mandatory_parameters = tuple()
    multi_valued_parameters = tuple()

    def __init__(self, name, params, parent=None):
        self.validate_parameters(params)

        self.name = name
        self.params = params
        self.parent = parent

    def _format_params(self):
        params_format = ''
        for param_name in self.params:
            param_values = self.params[param_name]
            # if is not multi-valued, is a list with 1 item.
            if not isinstance(param_values, (list, tuple)):
                param_values = [param_values]

            for param_value in param_values:
                params_format += '    %(param_name)s = %(param_value)s\n' % \
                                 {'param_name': param_name,
                                  'param_value': param_value}
        return params_format

    def _format_parent(self):
        """
        Formats configuration parent, if any.
        """
        if self.parent:
            return ': %(parent_name)s'.format({'parent_name': self.parent})
        else:
            return ''

    def format_output(self):
        """
        Formats this configuration into a string ready for sphinx.conf.
        """
        if not self.params:
            return ''

        return '%(type_name)s %(name)s %(parent_format)s\n{\n%(params_format)s}\n' % \
               {'type_name': self.type_name,
                'name': self.name,
                'parent_format': self._format_parent(),
                'params_format': self._format_params()}

    @classmethod
    def validate_parameters(cls, params):
        """
        Checks that all parameters `params` are valid for this configuration.
        """
        missing_parameters = set(cls.mandatory_parameters)

        for param_name in params:
            # check if it is valid.
            if param_name not in cls.valid_parameters:
                raise ImproperlyConfigured(
                    'Invalid parameter "{0}" for {1}. '
                    'See Sphinx documentation for "{1} configuration".'
                    .format(param_name, cls.type_name))

            # check if it is mandatory.
            if param_name in missing_parameters:
                missing_parameters.remove(param_name)

            # check if it is correctly typed.
            if param_name in cls.multi_valued_parameters:
                valid_types = (str, int, list, tuple)
            else:
                valid_types = (str, int)

            if not isinstance(params[param_name], valid_types):
                raise ImproperlyConfigured(
                    'Parameter "{0}" in {1} must be instance of '
                    '{2}.'.format(param_name, cls.type_name, valid_types))

            # check if each entry of the iterable is correctly typed.
            if isinstance(params[param_name], (list, tuple)):
                for entry in params[param_name]:
                    if not isinstance(entry, (str, int)):
                        raise ImproperlyConfigured(
                            'Item "{0}" of parameter "{1}" in {2} has wrong type.'
                            .format(entry, param_name, cls.type_name))

        if missing_parameters:
            raise ImproperlyConfigured(
                'Missing parameter(s) {0} for "{1}". '
                'See Sphinx documentation for {1} configuration.'
                .format(missing_parameters, cls.type_name))


class IndexConfiguration(Configuration):
    """
    Responsible for configuring a Sphinx source and index
    """
    type_name = 'index'
    valid_parameters = constants.index_parameters
    mandatory_parameters = constants.index_mandatory_parameters
    multi_valued_parameters = constants.index_multi_valued_parameters


class SourceConfiguration(Configuration):
    """
    Responsible for configuring a Sphinx source and index
    """
    type_name = 'source'
    valid_parameters = constants.source_parameters
    mandatory_parameters = constants.source_mandatory_parameters
    multi_valued_parameters = constants.source_multi_valued_parameters


class IndexerConfiguration(Configuration):
    """
    Responsible for configuring a Sphinx source and index
    """
    type_name = 'indexer'
    valid_parameters = constants.indexer_parameters

    def __init__(self, params):
        super(IndexerConfiguration, self).__init__('', params)


class SearchdConfiguration(Configuration):
    """
    Responsible for configuring a Sphinx source and index
    """
    type_name = 'searchd'
    valid_parameters = constants.searchd_parameters
    mandatory_parameters = constants.searchd_mandatory_parameters

    def __init__(self, params):
        super(SearchdConfiguration, self).__init__('', params)



from django.db.models import F, CharField, Value
from django.db.models.functions import Concat
from sphinxql import indexes, fields

from .models import Document


class DocumentIndex(indexes.Index):
    text = fields.IndexedString(model_attr='text')
    type_name = fields.IndexedString(model_attr='type__name')
    main_type_name = fields.IndexedString(model_attr='type__type__name')
    main_type_name1 = fields.IndexedString(model_attr=F('type__type__name'))
    type_name2 = fields.IndexedString(model_attr=Concat('type__type__name',
                                                        Value(" "), 'type__name',
                                                        output_field=CharField()))
    date = fields.Date(model_attr='type__date')

    class Meta:
        model = Document


class DocumentIndex1(indexes.Index):
    type_name2 = fields.IndexedString(model_attr=Concat('text', Value(" "),
                                                        'text',
                                                        output_field=CharField()))

    class Meta:
        model = Document


class DocumentIndex2(indexes.Index):
    name = fields.IndexedString(Concat(F('type__name'), Value(' '),
                                       output_field=CharField()))
    text = fields.IndexedString('text')

    class Meta:
        model = Document
        range_step = 10000



from .core.base import Match, Neg, Count, All
from .core.columns import Column, IdColumn, WeightColumn


import sphinxql.core.base


class C(sphinxql.core.base.SQLExpression):
    """
    The only element capable of resolving columns
    from strings.
    """
    def resolve_columns(self, index):
        """
        Returns the ``Field`` (that is, a Column) of an ``Index`` from its name.
        """
        if self._value == '@id':
            return IdColumn()
        elif self._value == '@relevance':
            return WeightColumn()

        try:
            return index.__dict__[self._value]
        except KeyError:
            raise KeyError('Field "%s" not found in "%s"' %
                           (self._value, index.__name__))


class _Infix:
    """
    Define an arbitrary operator.

    Minimalistic code from
    http://code.activestate.com/recipes/384122-infix-operators/
    """
    def __init__(self, function):
        self.function = function

    def __ror__(self, other):
        return _Infix(lambda x, self=self, other=other: self.function(other, x))

    def __or__(self, other):
        return self.function(other)


@_Infix
def And(one, other):
    return sphinxql.core.base.And(one, other)


@_Infix
def In(element, set):
    return sphinxql.core.base.In(element, set)


@_Infix
def NotIn(element, set):
    return sphinxql.core.base.NotIn(element, set)


@_Infix
def Between(element, set):
    return sphinxql.core.base.Between(element, set)

@_Infix
def NotBetween(element, set):
    return sphinxql.core.base.NotBetween(element, set)



import inspect
from django.utils import six


class IndexManager(object):

    def __init__(self, index=None, using=None, hints=None):
        self.index = index
        self._db = using
        self._hints = hints

    # Copy from BaseManager
    @classmethod
    def from_queryset(cls, queryset_class, class_name=None):
        if class_name is None:
            class_name = '%sFrom%s' % (cls.__name__, queryset_class.__name__)
        class_dict = {
            '_queryset_class': queryset_class,
        }
        class_dict.update(cls._get_queryset_methods(queryset_class))
        return type(class_name, (cls,), class_dict)

    # Copy from BaseManager
    @classmethod
    def _get_queryset_methods(cls, queryset_class):
        def create_method(name, method):
            def manager_method(self, *args, **kwargs):
                return getattr(self.get_queryset(), name)(*args, **kwargs)
            manager_method.__name__ = method.__name__
            manager_method.__doc__ = method.__doc__
            return manager_method

        new_methods = {}
        # Refs http://bugs.python.org/issue1785.
        predicate = inspect.isfunction if six.PY3 else inspect.ismethod
        for name, method in inspect.getmembers(queryset_class, predicate=predicate):
            # Only copy missing methods.
            if hasattr(cls, name):
                continue
            # Only copy public methods or methods with the attribute `queryset_only=False`.
            queryset_only = getattr(method, 'queryset_only', None)
            if queryset_only or (queryset_only is None and name.startswith('_')):
                continue
            # Copy the method onto the manager.
            new_methods[name] = create_method(name, method)
        return new_methods

    def get_queryset(self):
        """
        Returns a new QuerySet object.  Subclasses can override this method to
        easily customize the behavior of the Manager.
        """
        return self._queryset_class(index=self.index, using=self._db,
                                    hints=self._hints)



from collections import OrderedDict
from .configuration import indexes_configurator
from .exceptions import ImproperlyConfigured
from .fields import Field
from .query import SearchQuerySet
from .manager import IndexManager


class MetaIndex(type):

    # Ensures the fields are ordered when first created.
    @classmethod
    def __prepare__(mcs, name, bases):
        return OrderedDict()

    def __new__(mcs, name, bases, attrs):
        # exclude ``Index`` itself
        parents = [b for b in bases if isinstance(b, MetaIndex)]
        if not parents:
            return super(MetaIndex, mcs).__new__(mcs, name, bases, attrs)

        meta = attrs['Meta']

        model_meta = getattr(meta.model, '_meta', None)
        if not model_meta:
            raise ImproperlyConfigured('{index_name}.Meta.model "{model}" '
                                       'is not a Django model'.format(index_name=name,
                                                                      model=meta.model))

        # create new class
        new_class = super(MetaIndex, mcs).__new__(mcs, name, bases, dict(attrs))
        meta = new_class.Meta

        # populate Meta with fields
        meta.fields = []
        for field_name in attrs:
            # ignore non-fields
            if not isinstance(attrs[field_name], Field):
                continue
            field = attrs[field_name]

            # set field name and add it to meta
            field._value = field_name
            meta.fields.append(field)

        # register the field in our index configurator
        # so it is indexed by Sphinx.
        indexes_configurator.register(new_class)

        # managers
        has_any_manager = False
        for manager_name in attrs:
            # ignore non-managers
            if not isinstance(attrs[manager_name], IndexManager):
                continue

            # construct the manager from the queryset, init it with this index and
            # set it to the index
            manager = attrs[manager_name].from_queryset(SearchQuerySet)(new_class)
            setattr(new_class, manager_name, manager)
            has_any_manager = True

        if not has_any_manager:
            new_class.objects = IndexManager.from_queryset(SearchQuerySet)(new_class)

        return new_class


class Index(metaclass=MetaIndex):
    """
    Base class of all indexes.
    """

    class Meta:
        model = None

    @classmethod
    def build_name(cls):
        """
        Returns a string representing the index
        on the sphinx.conf.
        """
        meta = getattr(cls.Meta.model, '_meta', None)
        assert meta is not None
        return "%s_%s" % (meta.app_label, cls.__name__.lower())

    @classmethod
    def get_model_field(cls, model_attr):
        field, _, _, _ = cls.Meta.model._meta.get_field_by_name(model_attr)
        return field.get_attname_column()[1]



from django.utils.timezone import now

from sphinxql.core.base import DateTime, Date, Count, All
from sphinxql.query import Query
from sphinxql.sql import Match

from .indexes import DocumentIndex
from .models import Document

from tests import SphinxQLTestCase


class IndexTestCase(SphinxQLTestCase):

    def setUp(self):
        super(IndexTestCase, self).setUp()

        self.document = Document.objects.create(
            summary="This is a summary", text="What a nice text",
            date=now().date(),
            added_time=now().replace(microsecond=0),
            number=2, float=2.2, bool=True,
            unicode='câmara',
            slash='/summary')
        self.index()

        self.query = Query()
        self.query.fromm.append(DocumentIndex)

    def test_summary(self):
        result = list(self.query)[0][1]
        self.assertEqual(result, 'This is a summary')

    def test_text(self):
        result = list(self.query)[0][2]
        self.assertEqual(result, 'What a nice text')

    def test_date(self):
        result = list(self.query)[0][3]

        result = Date.to_python(result)
        self.assertEqual(result, self.document.date)

    def test_datetime(self):
        result = list(self.query)[0][4]

        result = DateTime.to_python(result)
        self.assertEqual(result, self.document.added_time)

    def test_number(self):
        result = list(self.query)[0][5]
        self.assertEqual(result, 2)

    def test_float(self):
        result = list(self.query)[0][6]
        self.assertEqual(result, 2.2)

    def test_bool(self):
        result = list(self.query)
        self.assertEqual(len(result), 1)

        result = result[0][7]
        self.assertEqual(result, 1)

    def test_unicode(self):
        self.query.where = Match('@unicode câmara')
        self.assertEqual(len(self.query), 1)

        self.query.where = Match('@unicode c mara')
        self.assertEqual(len(self.query), 0)

    def test_slash(self):
        """
        Issue #6: slashes must be escaped correctly.
        """
        self.query.where = Match('@slash /summary')

        self.assertTrue(r'\/' in str(self.query))
        self.assertEqual(len(self.query), 1)


class IndexSeveralTestCase(SphinxQLTestCase):

    def setUp(self):
        super(IndexSeveralTestCase, self).setUp()

        for x in range(1000):
            self.document = Document.objects.create(
                summary="This is a summary", text="What a nice text",
                date=now().date(),
                added_time=now(),
                number=x, float=2.2, bool=True,
                unicode='câmara',
                slash='/summary')
        self.index()

        self.query = Query()
        self.query.fromm.append(DocumentIndex)

    def test_range_query(self):
        self.query.select.append(Count(All()))
        self.assertEqual(list(self.query)[0][1], 1000)



from sphinxql import indexes, fields

from .models import Document


class DocumentIndex(indexes.Index):
    my_summary = fields.IndexedString(model_attr='summary')
    my_text = fields.IndexedString(model_attr='text')

    my_date = fields.Date(model_attr='date')
    my_added_time = fields.DateTime(model_attr='added_time')

    my_number = fields.Integer(model_attr='number')

    my_float = fields.Float(model_attr='float')

    my_bool = fields.Bool(model_attr='bool')

    unicode = fields.IndexedString(model_attr='unicode')

    slash = fields.IndexedString(model_attr='slash')

    class Meta:
        model = Document
        range_step = 100



from __future__ import unicode_literals

from django.core.management.base import BaseCommand

from sphinxql import configuration


class Command(BaseCommand):
    help = "Indexes your models."

    def add_arguments(self, parser):
        parser.add_argument(
            '--update',
            action='store_true',
            help='',
            default=True)

    def handle(self, **options):
        self.stdout.write('Started indexing')
        self.stdout.write('----------------')

        if options['update']:
            self.stdout.write(configuration.reindex())
        else:
            self.stdout.write(configuration.index())

        self.stdout.write('-----------------')
        self.stdout.write('Indexing finished')



import subprocess
import time

from .configurators import Configurator

indexes_configurator = Configurator()


def call_process(args, fail_silently=False):
    p = subprocess.Popen(args,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    p.wait()

    out = p.stdout.read().decode('UTF-8')
    if p.returncode != 0 and not fail_silently:
        raise Exception('\n\n' + out)
    return out


def index():
    return call_process(['indexer', '--all', '--config',
                         indexes_configurator.sphinx_file])


def reindex():
    out = call_process(['indexer', '--all', '--rotate', '--config',
                        indexes_configurator.sphinx_file])
    # it is not immediately available; wait a bit
    # see http://sphinxsearch.com/bugs/view.php?id=2350
    time.sleep(0.5)
    return out


def start(silent_fail=False):
    return call_process(['searchd', '--config',
                         indexes_configurator.sphinx_file], silent_fail)


def stop(silent_fail=False):
    return call_process(['searchd', '--stopwait', '--config',
                         indexes_configurator.sphinx_file], silent_fail)


def statistics():
    return call_process(['indextool', '--dumpheader', 'query_authorindex',
                         '--config', indexes_configurator.sphinx_file],
                        fail_silently=True)


def restart():
    stop(silent_fail=True)
    start()



from setuptools import setup, find_packages

setup(name='django-sphinxql',
      version='1.1.0',
      description='Sphinx search on Django',
      long_description=open('README.md').read(),
      author='Jorge C. Leitão',
      author_email='jorgecarleitao@gmail.com',
      packages=find_packages(),
      install_requires=['Django >= 1.8', 'pymysql'],
      url='https://github.com/jorgecarleitao/django-sphinxql',
      license='GPLv2',
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Environment :: Web Environment',
          'Framework :: Django',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
          'Operating System :: Unix',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Topic :: Utilities',
      ],
)



from collections import OrderedDict
from unittest import TestCase
from sphinxql import indexes

from sphinxql.configuration.configurations import IndexConfiguration
from sphinxql.configuration.configurators import add_source_conf_param
from sphinxql.exceptions import ImproperlyConfigured


class ConfiguratorTestCase(TestCase):
    def test_format(self):
        index_configurator = IndexConfiguration('test',
                                                OrderedDict([('source', 'test1'),
                                                             ('type', 'tes1'),
                                                             ('path', 'mypath')]))

        expected = "index test \n{\n    source = test1\n    type = tes1\n    path = mypath\n}\n"

        self.assertEqual(index_configurator.format_output(), expected)

    def test_multi_parameter_format(self):
        index_configurator = IndexConfiguration('test',
                                                OrderedDict([('source', 'test1'),
                                                             ('type', 'test1'),
                                                             ('rt_field', ['test1', 'test2']),
                                                             ('path', 'mypath')]))

        expected = "index test \n{\n" \
                   "    source = test1\n" \
                   "    type = test1\n" \
                   "    rt_field = test1\n" \
                   "    rt_field = test2\n" \
                   "    path = mypath\n" \
                   "}\n"

        self.assertEqual(index_configurator.format_output(), expected)

    def test_wrong_parameter(self):
        self.assertRaises(ImproperlyConfigured,
                          IndexConfiguration, 'test', {'source': 'test1',
                                                       'typERROR': 'tes1',
                                                       'path': 'mypath'})

    def test_missing_mandatory(self):
        self.assertRaises(ImproperlyConfigured,
                          IndexConfiguration, 'test', {'source': 'test1',
                                                       'type': 'tes1'})

    def test_wrong_source_param(self):
        self.assertRaises(ImproperlyConfigured, add_source_conf_param, {}, 'asdas', 1)

    def test_wrong_typed_parameter(self):
        with self.assertRaises(ImproperlyConfigured):
            IndexConfiguration('test',
                               {'source': 'test1',
                                'type': {'test1': 'test2'},
                                'path': 'mypath'})

    def test_wrong_typed_items(self):
        with self.assertRaises(ImproperlyConfigured):
            IndexConfiguration('test',
                               {'source': 'test1',
                                'type': 'test1',
                                'rt_field': [{'test1'}, 'test2'],
                                'path': 'mypath'})

    def test_not_a_model(self):
        with self.assertRaises(ImproperlyConfigured):
            class Index(indexes.Index):
                class Meta:
                    model = OrderedDict



import datetime

from django.db.models import Sum

from sphinxql.query import SearchQuerySet
from sphinxql.sql import C

from .indexes import DocumentIndex
from .models import Document

from tests import SphinxQLTestCase


class SearchQuerySetTestCase(SphinxQLTestCase):

    def setUp(self):
        super(SearchQuerySetTestCase, self).setUp()

        for x in range(1, 101):
            Document.objects.create(
                summary="This is a summary", text="What a nice text. "*x,
                date=datetime.date(2015, 2, 2) + datetime.timedelta(days=x),
                added_time=datetime.datetime(2015, 4, 4, 12, 12, 12) + datetime.timedelta(days=x),
                number=x*2)

        self.index()

        self.query = SearchQuerySet(DocumentIndex)

    def test_django_filter(self):
        query = self.query.filter(number__gt=190)
        self.assertEqual(len(query), 5)
        self.assertEqual(query[0].number, 192)

        q = query.search('@text text. What')
        self.assertEqual(len(q), 5)
        self.assertEqual(q[0].number, 200)

        q = query.search('@text text. What').search_order_by(-C('@relevance'))
        self.assertEqual(len(q), 5)
        self.assertEqual(q[0].number, 192)

    def test_search_filter(self):
        self.assertEqual(len(self.query), 100)
        # all except one should have this one
        self.assertEqual(len(self.query.search('@text "text. What"')), 99)

    def test_override_order_by(self):
        # order by in Django means first number is 200
        query = self.query.order_by('-id')
        self.assertEqual(query[0].number, 200)

        # don't override Django order means first number is 200
        query.search_mode = True
        query = query.search_order_by(C('@id'))
        self.assertEqual(query[0].number, 200)

        # clear Django ordering overrides it; order becomes @id.
        query = query.order_by()
        self.assertEqual(query[0].number, 2)

    def test_iter(self):
        q = self.query.search('@text What').search_order_by()
        i = 1
        for x in q:
            self.assertEqual(x.number, i*2)
            i += 1

    def test_change_queryset(self):

        q = self.query.search('@text What')
        q = q.filter(number__lte=40)
        self.assertEqual(len(q), 20)

        q = q.annotate(sum=Sum('number'))
        for x in q:
            self.assertTrue(hasattr(x, 'sum'))

    def test_data_is_cached(self):
        query = self.query.all()
        with self.assertNumQueries(1):
            list(query)
            list(query)

    def test_one_django_hit(self):
        query = self.query.all()
        with self.assertNumQueries(1):
            len(query)
            query[0:20]
            list(query)

    def test_search_override_default_ordering(self):
        self.assertEqual(self.query[0].number, 2)

        q = self.query.search('@text What')
        self.assertEqual(q[0].number, 200)

    def test_search_doesnt_override_ordering(self):
        q = self.query.search_order_by(C('number'))
        self.assertEqual(q[0].number, 2)

        q = q.search('@text What')
        self.assertEqual(q[0].number, 2)

    def test_django_ordering_override(self):
        # this adds a search_order_by
        q = self.query.search('@text What')
        q1 = self.query.order_by('number')
        self.assertEqual(q[0].number, 200)
        self.assertEqual(q1[0].number, 2)

        # Django order is not ignored
        q = q.order_by('number')
        self.assertEqual(q[0].number, 2)

    def test_django_order_not_overridden(self):
        q = self.query.search('@text What').search_order_by()
        self.assertEqual(q[0].number, 2)

        q = q.order_by('-number')
        self.assertEqual(q[0].number, 200)

    def test_manager(self):
        # test that `objects` work.
        q = DocumentIndex.objects.search('@text What')
        self.assertEqual(q[0].number, 200)

        self.assertEqual(DocumentIndex.objects.count(), 100)

        self.assertEqual(DocumentIndex.objects
                         .filter(number__in=[2, 4, 6]).count(), 3)

        self.assertEqual(DocumentIndex.other_objects.count(), 3)


class HighNumberSearchQuerySetTestCase(SphinxQLTestCase):
    """
    Test for the case with more than 1000 entries
    """
    def setUp(self):
        super(HighNumberSearchQuerySetTestCase, self).setUp()

        for x in range(1, 1005):
            Document.objects.create(
                summary="This is a summary", text="What a nice text.",
                date=datetime.date(2015, 2, 2) + datetime.timedelta(days=x),
                added_time=datetime.datetime(2015, 4, 4, 12, 12, 12) + datetime.timedelta(days=x),
                number=x)

        self.index()

        self.query = SearchQuerySet(DocumentIndex)

    def test_len(self):
        self.assertEqual(len(self.query.search('nice')), 1000)
        self.assertEqual(self.query.search('nice').count(), 1000)

    def test_django_len(self):
        query = self.query.filter(number__gte=95)  # exclude id in [1,94]
        self.assertEqual(len(query), 910)
        self.assertEqual(query.count(), 910)

        # since search results are ordered by id, they are ids 1-1000,
        # thus, the first 94 (id 1-94) + last 5 (id 1000-1005) are discarded
        self.assertEqual(len(query.search('@text nice')), 906)
        self.assertEqual(query.search('nice').count(), 906)

        query = query.search_order_by(-C('@id'))
        self.assertEqual(len(query.search('@text nice')), 910)
        self.assertEqual(query.search('nice').count(), 910)



from unittest import TestCase, expectedFailure
import datetime

from sphinxql.core.base import Function, Or
from sphinxql.sql import Column, And, In, NotIn, Between, NotBetween
from sphinxql.types import Integer, Bool, Date


class ExpressionTestCase(TestCase):

    def setUp(self):
        self.column = Column(Integer, 'test')
        self.column1 = Column(Integer, 'test1')

    def test_basic(self):
        self.assertEqual(self.column.type(), Integer)
        self.assertEqual(self.column.sql(), '`test`')

        self.column **= 2
        self.assertEqual(self.column.type(), Integer)
        self.assertEqual(self.column.sql(), 'POW(`test`, 2)')

    def test_plus(self):
        r = self.column + self.column1
        self.assertEqual(r.sql(), '`test` + `test1`')

        r = self.column + 2
        self.assertEqual(r.sql(), '`test` + 2')

        r = 2 + self.column
        self.assertEqual(r.sql(), '2 + `test`')

    def test_minus(self):
        r = self.column - 2
        self.assertEqual(r.sql(), '`test` - 2')

        r = -2 + self.column
        self.assertEqual(r.sql(), '-2 + `test`')

        r = -self.column + 2
        self.assertEqual(r.sql(), '-`test` + 2')

        r = 2 - self.column
        self.assertEqual(r.sql(), '2 - `test`')

    def test_times(self):
        r = self.column * 2
        self.assertEqual(r.sql(), '`test` * 2')

        r = 2 * self.column
        self.assertEqual(r.sql(), '2 * `test`')

    def test_divide(self):
        r = self.column / 2
        self.assertEqual(r.sql(), '`test` / 2')

        r = 2 / self.column
        self.assertEqual(r.sql(), '2 / `test`')

    def test_power(self):
        r = self.column ** 2
        self.assertEqual(r.sql(), 'POW(`test`, 2)')

        r = 2 ** self.column
        self.assertEqual(r.sql(), 'POW(2, `test`)')


class BooleansTestCase(TestCase):
    def setUp(self):
        self.column = Column(Integer, 'test')

    def test_equal(self):
        r = self.column == 2
        self.assertEqual(r.type(), Bool)
        self.assertEqual(r.sql(), '`test` = 2')

    def test_neq(self):
        r = self.column != 2
        self.assertEqual(r.type(), Bool)
        self.assertEqual(r.sql(), '`test` != 2')

    def test_neg(self):
        r = ~(self.column != 2)
        self.assertEqual(r.type(), Bool)
        self.assertEqual(r.sql(), 'NOT (`test` != 2)')

    def test_gt(self):
        r = self.column > 2
        self.assertEqual(r.type(), Bool)
        self.assertEqual(r.sql(), '`test` > 2')

    def test_ge(self):
        r = self.column >= 2
        self.assertEqual(r.type(), Bool)
        self.assertEqual(r.sql(), '`test` >= 2')

    def test_lt(self):
        r = self.column < 2
        self.assertEqual(r.type(), Bool)
        self.assertEqual(r.sql(), '`test` < 2')

    def test_le(self):
        r = self.column <= 2
        self.assertEqual(r.type(), Bool)
        self.assertEqual(r.sql(), '`test` <= 2')

    @expectedFailure
    def test_or(self):
        """
        OR not defined in SphinxQL yet.
        """
        r = (self.column <= 2) |Or| (self.column >= 2)
        self.assertEqual(r.type(), Bool)
        self.assertEqual(r.sql(), '(`test` <= 2) OR (`test` >= 2)')

    def test_and(self):
        r = (self.column <= 2) |And| (self.column >= 2)
        self.assertEqual(r.type(), Bool)
        self.assertEqual(r.sql(), '(`test` <= 2) AND (`test` >= 2)')

    def test_two_and(self):
        r = (self.column <= 2) |And| (self.column >= 2) |And| (self.column == 2)
        self.assertEqual(r.type(), Bool)
        self.assertEqual(r.sql(), '((`test` <= 2) AND (`test` >= 2)) AND (`test` = 2)')

    def test_in(self):
        r = self.column |In| (2, 3, 4, 5)
        self.assertEqual(r.type(), Bool)
        self.assertEqual(r.sql(), '`test` IN (2, 3, 4, 5)')

    def test_not_in(self):
        r = self.column |NotIn| (2, 3, 4, 5)
        self.assertEqual(r.type(), Bool)
        self.assertEqual(r.sql(), '`test` NOT IN (2, 3, 4, 5)')

    def test_between(self):
        r = self.column |Between| (2, 3)
        self.assertEqual(r.type(), Bool)
        self.assertEqual(r.sql(), '`test` BETWEEN 2 AND 3')

    def test_not_between(self):
        r = self.column |NotBetween| (2, 3)
        self.assertEqual(r.type(), Bool)
        self.assertEqual(r.sql(), '`test` NOT BETWEEN 2 AND 3')


class OtherTestCase(TestCase):
    def setUp(self):
        self.column = Column(Integer, 'test')

    def test_float(self):

        r = self.column + 2.2

        self.assertEqual(r.type(), Integer)
        self.assertEqual(r.as_sql(), '`test` + 2.200000')

    def test_bool(self):

        r = self.column + True

        self.assertEqual(Bool(True).type(), Bool)
        self.assertEqual(r.type(), Integer)
        self.assertEqual(r.as_sql(), '`test` + TRUE')

    def test_dates(self):

        r = Date(datetime.date(2014, 2, 2)) > datetime.date(2014, 3, 2)

        self.assertEqual(r.type(), Bool)
        self.assertEqual(r.as_sql(), '1391299200 > 1393718400')

    def test_datetimes(self):

        r = Date(datetime.datetime(2014, 2, 2, 12, 12, 12)) > datetime.datetime(2014, 2, 2, 12, 12, 13)

        self.assertEqual(r.type(), Bool)
        self.assertEqual(r.as_sql(), '1391343132 > 1391343133')

    def test_wrong_type(self):
        self.assertRaises(TypeError, Date.__lt__, Date(datetime.datetime(2014, 2, 2)), And)

    def test_wrong_function_arguments(self):
        self.assertRaises(IndexError, Function, [1, 2])



from django.db import models


class MainType(models.Model):
    name = models.CharField(max_length=200)


class Type(models.Model):
    name = models.CharField(max_length=200)
    type = models.ForeignKey(MainType)
    date = models.DateField()


class Document(models.Model):
    text = models.TextField()
    type = models.ForeignKey(Type)



from django.db import models


class Document(models.Model):
    summary = models.CharField(max_length=254)
    text = models.TextField()

    date = models.DateField()
    added_time = models.DateTimeField()

    number = models.IntegerField()

    float = models.FloatField()

    bool = models.BooleanField(default=False)

    unicode = models.CharField(max_length=254)

    slash = models.CharField(max_length=20)



from collections import OrderedDict

import django.db.models.query

from .core.query import Query
from .core import base
from .core.lookups import LOOKUP_SEPARATOR, parse_lookup
from sphinxql.exceptions import NotSupportedError
from .types import Bool
from .sql import Match, And, Neg, C, Column, All, Count


class QuerySet(object):

    def __init__(self, index):
        self._index = index
        self.query = Query()
        self.query.fromm.append(index)

        # Sphinx: there can be only one match per query.
        # This is a global constraint on queries, we keep it here.
        self._match = ''

        self._result_cache = None
        self._fetch_cache = None

        self._set_default_fields(self.query)

    def _fetch_raw(self):
        """
        Fetches by hitting Sphinx
        """
        if self._fetch_cache is not None:
            return self._fetch_cache

        self._fetch_cache = list(self._get_query())
        return self._fetch_cache

    def _get_query(self):
        """
        Returns a copy of the query exactly prior to hit db.
        """
        clone = self.query.clone()
        if self._match:
            clone.where = self._add_condition(clone.where, Match(self._match))
        return clone

    def _parsed_results(self):
        """
        Hits Sphinx and parses the results into indexes instances.
        """
        for result in self._fetch_raw():
            instance = self._index()

            setattr(instance, 'id', result[0])
            i = 1  # 1 is id
            for field in self._index.Meta.fields:
                if field.is_attribute:
                    setattr(instance, field.name, field.type().to_python(result[i]))
                    i += 1

            yield instance

    def __iter__(self):
        if self.query.limit is None:
            raise IndexError('Sphinx does not support unbounded iterations '
                             'over the results.')
        return self._parsed_results()

    def __len__(self):
        if self._fetch_cache is not None:
            return len(self._fetch_cache)
        return self.count()

    def __getitem__(self, item):
        if not isinstance(item, (slice, int)):
            raise TypeError
        if isinstance(item, slice):
            if item.stop is None:
                raise NotSupportedError('Sphinx does not support '
                                        'unbounded slicing.')

        if isinstance(item, slice):
            offset = item.start or 0
            count = item.stop - offset

            clone = self.clone()
            clone.query.limit = (offset, count)
            return list(clone)
        else:
            offset = item
            count = 1
            clone = self.clone()
            clone.query.limit = (offset, count)
            return list(clone)[0]

    def all(self):
        return self

    def count(self):
        q = self._get_query()
        q.select.clear()
        q.select.append(Count(All()))

        result = list(q)
        if result:
            # first row, second entry (first entry is row's `id`)
            return result[0][1]
        else:
            return 0

    def filter(self, *conditions, **lookups):
        clone = self.clone()

        conditions = list(conditions)
        for lookup in lookups:
            condition = parse_lookup(lookup, lookups[lookup])
            conditions.append(condition)

        for condition in conditions:
            assert isinstance(condition, base.SQLExpression)
            condition = condition.resolve_columns(self._index)
            assert condition.type() == Bool
            clone.query.where = self._add_condition(clone.query.where, condition)

        return clone

    def search(self, *extended_queries):
        clone = self.clone()
        if clone._match == '':
            clone._match = ' '.join(list(extended_queries))
        else:
            clone._match = ' '.join([clone._match] + list(extended_queries))
        return clone

    def order_by(self, *args):
        """
        Only accepts Neg, C, Columns
        """
        clone = self.clone()

        if not args:
            clone.query.order_by.clear()
            return clone

        for arg in args:
            # parse string
            if isinstance(arg, str):
                if LOOKUP_SEPARATOR in arg:
                    raise NotImplementedError('Django-SphinxQL does not support '
                                              'lookups in order by.')
                if arg[0] == '-':
                    arg = Neg(C(arg[1:]))
                else:
                    arg = C(arg)

            # parse negation of column
            assert isinstance(arg, (Neg, C, Column))
            ascending = True
            if isinstance(arg, Neg):
                ascending = False
                arg = arg.value[0]
                assert isinstance(arg, (C, Column))
            if isinstance(arg, C):
                column = arg.resolve_columns(clone._index)
            else:
                column = arg

            clone.query.order_by.append(column, ascending=ascending)

        return clone

    def _set_default_fields(self, query):
        fields = self._index.Meta.fields

        query.select.clear()
        for field in fields:
            if field.is_attribute:
                query.select.append(field)

    @staticmethod
    def _add_condition(where, condition):
        if where is None:
            where = condition
        else:
            where = where |And| condition
        return where

    def clone(self):
        clone = QuerySet(self._index)
        clone._match = self._match
        clone.query = self.query.clone()
        return clone


class SearchQuerySet(django.db.models.query.QuerySet):
    """
    A queryset to translate search results into Django models.
    """
    max_search_count = 1000

    def __init__(self, index, query=None, using=None, hints=None):
        super(SearchQuerySet, self).__init__(index.Meta.model, query, using, hints=hints)
        self._index = index
        self._sphinx_queryset = QuerySet(index)

        self._result_cache = None
        self.search_mode = False

    def search_filter(self, *conditions, **lookups):
        clone = self._clone()
        clone._sphinx_queryset = self._sphinx_queryset.filter(*conditions,
                                                              **lookups)
        return clone

    def search(self, *extended_queries):
        clone = self._clone()
        clone.search_mode = True
        clone._sphinx_queryset = clone._sphinx_queryset.search(*extended_queries)
        if not clone._sphinx_queryset.query.order_by:
            clone = clone.search_order_by(C('@relevance'))
        return clone

    def search_order_by(self, *columns):
        clone = self._clone()
        clone._sphinx_queryset = clone._sphinx_queryset.order_by(*columns)
        return clone

    def _has_explicit_ordering(self):
        """
        A weaker version of ``ordered`` that ignores default ordering and
        Meta.ordering.
        """
        if self.query.extra_order_by or self.query.order_by:
            return True
        return False

    def _annotated_models(self):
        """
        Returns the models annotated with `search_result`. Uses `_result_cache`.
        """
        if self._result_cache is not None:
            return self._result_cache

        # hit Sphinx: ordered results with index objects populated
        indexes = OrderedDict([(index_obj.id, index_obj)
                               for index_obj in
                               self._sphinx_queryset[:self.max_search_count]])

        # hit Django: ordered results with model objects populated
        clone = self._get_query(indexes.keys())
        models = OrderedDict([(obj.id, obj) for obj in clone])

        # exclude objects excluded by Django query
        # annotate models with search_result.
        self._result_cache = []
        if self._sphinx_queryset.query.order_by and \
                not self._has_explicit_ordering():
            for id in indexes:
                if id in models:
                    # annotate `search_result`
                    models[id].search_result = indexes[id]
                    self._result_cache.append(models[id])
        else:
            for id in models:
                # annotate `search_result`
                models[id].search_result = indexes[id]
                self._result_cache.append(models[id])
        return self._result_cache

    def _get_query(self, id_list=None):
        """
        Returns a Django queryset restricted to the ids in `id_list`.
        If `id_list` is None, hits Sphinx to retrieve it.
        """
        if id_list is None:
            id_list = [index_obj.id for index_obj in
                       self._sphinx_queryset[:self.max_search_count]]
        clone = self._clone()
        clone = clone.filter(pk__in=id_list)
        clone.search_mode = False
        return clone

    def __iter__(self):
        if self.search_mode:
            return iter(self._annotated_models())
        return super(SearchQuerySet, self).__iter__()

    def __len__(self):
        if self.search_mode:
            return len(self._get_query())
        return super(SearchQuerySet, self).__len__()

    def __getitem__(self, item):
        if self.search_mode:
            return self._annotated_models()[item]
        return super(SearchQuerySet, self).__getitem__(item)

    def count(self):
        if self.search_mode:
            return self._get_query().count()
        return super(SearchQuerySet, self).count()

    def _clone(self, klass=None, setup=False, **kwargs):
        ## almost-copy of original _clone:
        if klass is None:
            klass = self.__class__
        query = self.query.clone()
        if self._sticky_filter:
            query.filter_is_sticky = True
        # next line is different: first argument is index instead of model
        c = klass(index=self._index, query=query, using=self._db)
        c._for_write = self._for_write
        c._prefetch_related_lookups = self._prefetch_related_lookups[:]
        c._known_related_objects = self._known_related_objects
        c.__dict__.update(kwargs)
        if setup and hasattr(c, '_setup_query'):
            c._setup_query()

        # sphinx related
        c.search_mode = self.search_mode
        c._sphinx_queryset = self._sphinx_queryset.clone()
        return c



from . import base
from .. import sql

LOOKUPS = {'eq': base.Equal,
           'neq': base.NotEqual,
           'lt': base.LessThan,
           'gt': base.GreaterThan,
           'lte': base.LessEqualThan,
           'gte': base.GreaterEqualThan,
           'in': base.In,
           'nin': base.NotIn,
           'range': base.Between,
           'nrange': base.NotBetween}

LOOKUP_SEPARATOR = '__'


def parse_lookup(lhs, rhs):
    assert lhs
    parts = lhs.split(LOOKUP_SEPARATOR)

    if len(parts) > 2:
        raise NotImplementedError('Currently Django-SphinxQL only accepts'
                                  'lookups having up to 1 \'__\'.')

    if len(parts) == 1:
        parts.append('eq')

    if parts[0] == 'id':
        column = sql.C('@id')
    else:
        column = sql.C(parts[0])
    lookup = parts[1]

    try:
        operation = LOOKUPS[lookup]
    except KeyError:
        raise KeyError('Lookup \'{0}\' not valid. '
                       'Check documentation on available lookups.'
                       .format(parts[1]))

    if lookup not in ('in', 'nin', 'range', 'nrange'):
        rhs = base.convert(rhs)

    return operation(column, rhs)




class SphinxError(Exception):
    """
    general exception of Sphinx so users can catch any Sphinx error.
    """
    pass


class ImproperlyConfigured(SphinxError):
    """somehow improperly configured"""
    pass


class NotSupportedError(SphinxError):
    """Something not support by Sphinx"""
    pass
